@@ -418,14 +418,10 @@ static long _mpq_pythonhash(mpq_t the_rat)
     mpq_t rat;
     mpq_init(rat);
     mpq_set(rat, the_rat);
-    mpq_canonicalize(rat);
     long n = _mpz_pythonhash(mpq_numref(rat));
     long d = _mpz_pythonhash(mpq_denref(rat));
-    if (d != 1L) {
-        n = n^d;
-        if (n == -1L)
-            n = -2L;
-        }
+    if (d != 1L)
+        n = n + (d-1) * 7461864723258187525;
     mpq_clear(rat);
     return n;
 }
@@ -517,7 +513,7 @@ numeric::numeric(PyObject* o, bool force_py) : basic(&numeric::tinfo_static) {
         }
 
         t = PYOBJECT;
-        hash = PyObject_Hash(o);
+        hash = (long)PyObject_Hash(o);
         if (hash == -1 && PyErr_Occurred()) {
             // error is thrown on first hash request
             PyErr_Clear();
@@ -661,7 +657,7 @@ inherited(n, sym_lst) {
                         if (PyErr_Occurred()) {
                                 throw (std::runtime_error("archive error: caught exception in py_loads"));
                         }
-                        hash = PyObject_Hash(v._pyobject);
+                        hash = (long)PyObject_Hash(v._pyobject);
                         if (hash == -1 && PyErr_Occurred()) {
                             PyErr_Clear();
                             is_hashable = false;
@@ -922,15 +918,15 @@ bool numeric::is_equal_same_type(const basic &other) const {
         return this->is_equal(o);
 }
 
-unsigned numeric::calchash() const {
+long numeric::calchash() const {
         switch (t) {
                 case DOUBLE:
-                        return (unsigned int) v._double;
+                        return (long) v._double;
                 case MPZ:
                 case MPQ:
                 case PYOBJECT:
                         if (is_hashable)
-                            return (unsigned int) hash;
+                            return hash;
                         throw (std::runtime_error("Python object not hashable"));
                 default:
                         stub("invalid type: ::hash() type not handled");
@@ -974,7 +970,6 @@ const numeric numeric::add(const numeric &other) const {
                         mpq_t bigrat;
                         mpq_init(bigrat);
                         mpq_add(bigrat, v._bigrat, other.v._bigrat);
-                        mpq_canonicalize(bigrat);
                         return bigrat;
                 case PYOBJECT:
                         return PyNumber_Add(v._pyobject, other.v._pyobject);
@@ -1004,7 +999,6 @@ const numeric numeric::sub(const numeric &other) const {
                         mpq_t bigrat;
                         mpq_init(bigrat);
                         mpq_sub(bigrat, v._bigrat, other.v._bigrat);
-                        mpq_canonicalize(bigrat);
                         return bigrat;
                 case PYOBJECT:
                         return PyNumber_Subtract(v._pyobject, other.v._pyobject);
@@ -1034,7 +1028,6 @@ const numeric numeric::mul(const numeric &other) const {
                         mpq_t bigrat;
                         mpq_init(bigrat);
                         mpq_mul(bigrat, v._bigrat, other.v._bigrat);
-                        mpq_canonicalize(bigrat);
                         return bigrat;
                 case PYOBJECT:
                         return PyNumber_Multiply(v._pyobject, other.v._pyobject);
@@ -1074,14 +1067,12 @@ const numeric numeric::div(const numeric &other) const {
                                 mpq_set_z(obigrat, other.v._bigint);
                                 mpq_div(bigrat, bigrat, obigrat);
                                 mpq_clear(obigrat);
-                                mpq_canonicalize(bigrat);
                                 return bigrat;
                         }
                 case MPQ:
                                 mpq_t bigrat;
                                 mpq_init(bigrat);
                                 mpq_div(bigrat, v._bigrat, other.v._bigrat);
-                                mpq_canonicalize(bigrat);
                                 return bigrat;
 
                 case PYOBJECT:
@@ -1871,6 +1862,7 @@ PyObject* numeric::to_pyobject() const {
                         mpq_t bigrat;
                         mpq_init(bigrat);
                         mpq_set(bigrat, v._bigrat);
+                        mpq_canonicalize(bigrat);
                         o = py_funcs.py_rational_from_mpq(bigrat);
                         mpq_clear(bigrat);
                         return o;
@@ -2028,6 +2020,10 @@ const numeric numeric::zeta() const {
         PY_RETURN(py_funcs.py_zeta);
 }
 
+const numeric numeric::stieltjes() const {
+        PY_RETURN(py_funcs.py_stieltjes);
+}
+
 const numeric numeric::exp() const {
         PY_RETURN(py_funcs.py_exp);
 }
@@ -2490,6 +2486,10 @@ const numeric zeta(const numeric &x) {
         return x.zeta();
 }
 
+const numeric stieltjes(const numeric &x) {
+        return x.stieltjes();
+}
+
 /** The Gamma function.
  *  Use the Lanczos approximation. If the coefficients used here are not
  *  sufficiently many or sufficiently accurate, more can be calculated
